#ifndef FRAME_CPP_INTERFACE__FRAME_C_INTERNAL_HH
#define FRAME_CPP_INTERFACE__FRAME_C_INTERNAL_HH

#ifndef __cplusplus
#error This file needs to be compiled using C++
#endif /* __cplusplus */

#include <cstdlib>

#include "general/unordered_map.hh"
#include "general/mutexlock.hh"

#include "framec/FrameC.h"

namespace FrameC
{
  extern const char* UnknownError;

  template< typename T >
  void Deposit( T Value );

  template< typename T >
  void Validate( T Value );

  template< typename T >
  void Withdraw( T Value );

  void Set( FrameCError** Error );

  void Set( FrameCError** Error,
	    framec_errno_type Errno,
	    const std::string& Message );

  class Pointers
  {
  public:
    enum pointer_type {
      STREAM,
      STREAM_FILE_INPUT,
      STREAM_FILE_OUTPUT,
      FR_DETECTOR,
      FR_FRAME,
      FR_TOC
    };

    Pointers( );

    void Deposit( void* Ptr, pointer_type Type );

    void Validate( void* Ptr, pointer_type Type );

    void Withdraw( void* Ptr );

    template< int T >
    static const char* Name( );

  private:
    struct pointer_info {
      pointer_type	s_type;
    };

    struct hash_func {
      inline size_t operator( )( const void* Key ) const
      {
	return reinterpret_cast< size_t >( Key );
      }
    };

    typedef General::unordered_map< void*, pointer_info,
				    hash_func >  pointer_container_type;
    MutexLock::lock_type	m_baton;

    pointer_container_type	m_pointers;
  };

  extern Pointers	PointerHash;

  template< typename T >
  Pointers::pointer_type PointerType( T* );

  template< typename T >
  inline void
  Deposit( T* Value )
  {
    PointerHash.Deposit( Value, PointerType( Value ) );
  }

  template< typename T >
  inline void
  Validate( T* Value )
  {
    PointerHash.Validate( Value, PointerType( Value ) );
  }

  template< typename T >
  inline void
  Withdraw( T* Value )
  {
    Validate( Value );
    PointerHash.Withdraw( Value );
    delete Value;
  }

  inline void
  Set( FrameCError** Error )
  {
    if ( *Error )
    {
      free( (void*)((*Error)->s_message) );
      free( *Error );
    }
    *Error = (FrameCError*)NULL;
  }

  inline void
  Set( FrameCError** Error,
       framec_errno_type Errno,
       const std::string& Message )
  {
    FrameCError* retval = (FrameCError*)std::malloc( sizeof( FrameCError ) );

    retval->s_errno = Errno;
    retval->s_message = strdup( Message.c_str( ) );

    *Error = retval;
  }
}

#endif /* FRAME_CPP_INTERFACE__FRAME_C_INTERNAL_HH */
